
import asyncio
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, List


class Call:
    def __init__(self, *args: Any, **kwargs: Any):
        self.args = args
        self.kwargs = kwargs


@dataclass
class _BatchItem:
    call: Call
    result: Any


async def run_async_batch(
    func: Callable[..., Awaitable[Any]],
    calls: List[Call],
    max_retry_num: int = 2,
    max_iter_fail_ratio: float = 0.5,
) -> List[Any]:
    ''' Execute async operations concurrently.
        Args:
            func: async funcion
            calls: batch args, each element gives the parameter of an operation.
            max_retry_num: max retry number.
            max_iter_fail_ratio: Allowed max failure ratio in each retry.
            
        Returns:
            Results for each batch item.
    '''
    ex = Exception('unexpected error')
    call_batch = [_BatchItem(call=c, result=None) for c in calls]
    batch = call_batch[:]
    for _ in range(max_retry_num + 1):
        results = await asyncio.gather(
            *[func(*item.call.args, **item.call.kwargs) for item in batch],
            return_exceptions=True,
        )
        assert len(batch) == len(results)
        batch_retry = []
        for item, result in zip(batch, results):
            if isinstance(result, Exception):
                batch_retry.append(item)
                ex = result
            else:
                item.result = result

        if batch_retry:
            if len(batch_retry) > len(batch) * max_iter_fail_ratio:
                raise Exception(f'too high failure ratio for async batch: {len(batch_retry)} / {len(batch)}') from ex
            else:
                batch = batch_retry
        else:
            return [item.result for item in call_batch]
        
    raise Exception('too many retries for async batch') from ex




from __future__ import annotations
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
import asyncio
from threading import RLock
import time
from typing import Dict, List, Optional, Tuple, Union
import uuid
from azure.core import MatchConditions
from azure.core.credentials import TokenCredential
from azure.cosmos import CosmosClient, PartitionKey
from azure.cosmos import aio
from azure.cosmos.container import ContainerProxy
from azure.cosmos.exceptions import (
    CosmosAccessConditionFailedError,
    CosmosResourceNotFoundError,
    CosmosResourceExistsError,
)
from .common import (
    CONTAINER_COMMIT, CONTAINER_META, DATABASE_DYNCONF, PARTITION_KEY_DATA,
    CommitRecord, DataRecord, MetaRecord,
)
from .utils import run_async_batch, Call


# To exclude fields when update to remote cosmos.
FIELDS_EXCLUDE = {'etag', 'ts'}


class ItemExistsException(Exception):
    pass


##
## abstract
##

class DataAccessProxy(ABC):
    @abstractmethod
    def try_get_meta(self, app: str) -> Optional[MetaRecord]:
        ''' Get meta record by app.
            Args:
                app: dynamic config application name
            Returns:
                Got meta record, or None if not found.
        '''
        pass
    
    @abstractmethod
    def create_meta(self, meta: MetaRecord) -> MetaRecord:
        ''' Create meta record.
            Args:
                meta: meta record to create
            Returns:
                Created meta record.
            Raises:
                If meta record already exists, raise ItemExistsException.
        '''
        pass
    
    @abstractmethod
    def update_meta_if_not_modified(self, meta: MetaRecord, etag: str) -> Optional[MetaRecord]:
        ''' Update meta record if given etag matched with remote _etag.
            Args:
                meta: meta record to update
                etag: etag to match
            Returns:
                Updated meta record, or None if etag conflicts.
            Raises:
                If meta record not exists, raise an exception.
        '''
        pass

    @abstractmethod
    def create_commit(self, commit: CommitRecord) -> CommitRecord:
        ''' Create commit record.
            Args:
                commit: commit record to create.
            Returns:
                Created commit record.
            Raises:
                If commit record already exists, raise ItemExistsException.
        '''
        pass
    
    @abstractmethod
    def update_commit(self, commit: CommitRecord) -> CommitRecord:
        ''' Update commit record.
            Args:
                commit: commit record to update.
            Returns:
                Updated commit record.
            Raises:
                If commit record not exists, raise an exception.
        '''
        pass
    
    @abstractmethod
    def batch_delete_commits(self, commits: List[CommitRecord]):
        ''' Delete multiple commit records. Ignore non-existed.
            No requirements for atomic/trasaction.
            Args:
                commits: commit records to delete.
        '''
        pass
    
    @abstractmethod
    def query_commits_by_seq_range(self, app: str, seq_begin: int, seq_end: int) -> List[CommitRecord]:
        ''' Query commit records by sequence range.
            Args:
                app: dynamic config application name.
                seq_begin: begin of sequence range, inclusive.
                seq_end: end of sequence range, inclusive.
            Returns:
                List of queried commit records.
        '''
        pass
    
    @abstractmethod
    def query_data_items_by_commit_ids(self, app: str, commit_ids: List[str]) -> List[DataRecord]:
        ''' Get all the data item records belonging to given commits.
            Args:
                app: dynamic config application name.
                commit_ids: commit ids to query.
            Returns:
                Queried data item records.
        '''
        pass

    @abstractmethod
    def batch_upsert_data_items(self, app: str, data_items: List[DataRecord]):
        ''' Create or update for multiple data item records.
            No requirements for atomic/trasaction.
            Args:
                app: dynamic config application name.
                data_items: data item records to be upserted.
        '''
        pass
    
    @abstractmethod
    def batch_delete_data_items(self, app: str, data_items: List[DataRecord]):
        ''' Delete multiple data item records. Ignore non-existed.
            No requirements for atomic/trasaction.
            Args:
                app: dynamic config application name.
                data_items: data item records to be deleted.
        '''
        pass
    

##
## stubbed
##

class StubbedDataAccessProxy(DataAccessProxy):
    def __init__(self):
        self._metas: Dict[str, MetaRecord] = {}
        self._commits: Dict[str, CommitRecord] = {}
        self._data_items: Dict[Tuple[str, str], DataRecord] = {}  # {(app, id): record}
        self._lock = RLock()


    def try_get_meta(self, app: str) -> Optional[MetaRecord]:
        meta = self._metas.get(app)
        if meta is None:
            return meta
        else:
            return meta.copy(deep=True)


    def create_meta(self, meta: MetaRecord) -> MetaRecord:
        new = meta.copy(deep=True)
        new.etag = str(uuid.uuid4())
        new.ts = int(time.time())
        with self._lock:
            if new.id in self._metas:
                raise ItemExistsException(f'meta "{new.id}" already exists')
            self._metas[new.id] = new
        return new.copy(deep=True)
            
    
    def update_meta_if_not_modified(self, meta: MetaRecord, etag: str) -> Optional[MetaRecord]:
        new = meta.copy(deep=True)
        new.etag = str(uuid.uuid4())
        new.ts = int(time.time())
        with self._lock:
            old = self._metas.get(meta.id)
            if old is None:
                raise Exception('meta not exists')
            if old.etag != etag:
                return None
            self._metas[new.id] = new
        return new.copy(deep=True)
            

    def create_commit(self, commit: CommitRecord) -> CommitRecord:
        new = commit.copy(deep=True)
        new.etag = str(uuid.uuid4())
        new.ts = int(time.time())
        with self._lock:
            if new.id in self._commits:
                raise ItemExistsException(f'commit "{new.id}" already exists')
            self._commits[new.id] = new
        return new.copy(deep=True)
    
    
    def update_commit(self, commit: CommitRecord) -> CommitRecord:
        new = commit.copy(deep=True)
        new.etag = str(uuid.uuid4())
        new.ts = int(time.time())
        with self._lock:
            if new.id not in self._commits:
                raise Exception('commit not exists')
            self._commits[new.id] = new
        return new.copy(deep=True)

    
    def batch_delete_commits(self, commits: List[CommitRecord]):
        with self._lock:
            for commit in commits:
                if commit.id in self._commits:
                    self._commits.pop(commit.id)


    def query_commits_by_seq_range(self, app: str, seq_begin: int, seq_end: int) -> List[CommitRecord]:
        commits: List[CommitRecord] = []
        with self._lock:
            for commit in self._commits.values():
                if commit.app == app and seq_begin <= commit.seq <= seq_end:
                    commits.append(commit)
        return [c.copy(deep=True) for c in commits]
    
    
    def query_data_items_by_commit_ids(self, app: str, commit_ids: List[str]) -> List[DataRecord]:
        commit_idset = set(commit_ids)
        ditems: List[DataRecord] = []
        with self._lock:
            for (_app, _id), ditem in self._data_items.items():
                if _app == app and ditem.commitId in commit_idset:
                    ditems.append(ditem)
        return [ditem.copy(deep=True) for ditem in ditems]
    
    
    def batch_upsert_data_items(self, app: str, data_items: List[DataRecord]):
        with self._lock:
            for ditem in data_items:
                new = ditem.copy(deep=True)
                new.etag = str(uuid.uuid4())
                new.ts = int(time.time())
                self._data_items[(app, new.id)] = new


    def batch_delete_data_items(self, app: str, data_items: List[DataRecord]):
        with self._lock:
            for ditem in data_items:
                key = (app, ditem.id)
                if key in self._data_items:
                    self._data_items.pop(key)



##
## Cosmos Proxy
##

class CosmosDataAccessProxy(DataAccessProxy):
    ''' Designed for DynamicConfigConsumer.
        Cosmos synchronous client is not good at batch operations, some 
        batch functions are not implemented. Batch functions are only 
        required in DynamicConfigWriter.
    '''
    def __init__(self, client: CosmosClient, dbname: str = DATABASE_DYNCONF):
        ''' Args:
                client: cosmos client.
                dbname: database name. Default: "dynconfig".
        '''
        self._dbname = dbname
        self._client = client
        
        self._proxy_db = self._client.get_database_client(self._dbname)
        self._proxy_meta = self._proxy_db.get_container_client(CONTAINER_META)
        self._proxy_commit = self._proxy_db.get_container_client(CONTAINER_COMMIT)
        self._proxy_data_dict: Dict[str, ContainerProxy] = {}


    def _get_proxy_data(self, app: str) -> ContainerProxy:
        proxy_data = self._proxy_data_dict.get(app)
        if proxy_data is not None:
            return proxy_data
        
        # get container proxy
        assert app not in [CONTAINER_META, CONTAINER_COMMIT]
        proxy_data = self._proxy_db.get_container_client(app)

        # set
        self._proxy_data_dict[app] = proxy_data
        return proxy_data


    def try_get_meta(self, app: str) -> Optional[MetaRecord]:
        try:
            resp = self._proxy_meta.read_item(app, partition_key=app)
            return MetaRecord.parse_obj(resp)
        except CosmosResourceNotFoundError:
            return None
        
    
    def create_meta(self, meta: MetaRecord) -> MetaRecord:
        try:
            resp = self._proxy_meta.create_item(meta.dict(exclude=FIELDS_EXCLUDE))
            return MetaRecord.parse_obj(resp)
        except CosmosResourceExistsError as ex:
            raise ItemExistsException(f'meta "{meta.id}" already exists') from ex
    

    def update_meta_if_not_modified(self, meta: MetaRecord, etag: str) -> Optional[MetaRecord]:
        try:
            resp = self._proxy_meta.replace_item(
                meta.id,
                meta.dict(exclude=FIELDS_EXCLUDE),
                etag=etag,
                match_condition=MatchConditions.IfNotModified,
            )
            return MetaRecord.parse_obj(resp)
        except CosmosAccessConditionFailedError:
            return None


    def create_commit(self, commit: CommitRecord) -> CommitRecord:
        try:
            resp = self._proxy_commit.create_item(commit.dict(exclude=FIELDS_EXCLUDE))
            return CommitRecord.parse_obj(resp)
        except CosmosResourceExistsError as ex:
            raise ItemExistsException(f'commit "{commit.id}" already exists') from ex
    

    def update_commit(self, commit: CommitRecord) -> CommitRecord:
        resp = self._proxy_commit.replace_item(commit.id, commit.dict(exclude=FIELDS_EXCLUDE))
        return CommitRecord.parse_obj(resp)
    

    def batch_delete_commits(self, commits: List[CommitRecord]):
        raise NotImplementedError
    

    def query_commits_by_seq_range(self, app: str, seq_begin: int, seq_end: int) -> List[CommitRecord]:
        cursor = self._proxy_commit.query_items(
            'SELECT * FROM c WHERE c.seq between @seqBegin and @seqEnd',
            parameters=[
                dict(name="@seqBegin", value=seq_begin),
                dict(name="@seqEnd", value=seq_end),
            ],
            partition_key=app,
        )
        return [CommitRecord.parse_obj(item) for item in cursor]
    

    def query_data_items_by_commit_ids(self, app: str, commit_ids: List[str]) -> List[DataRecord]:
        proxy_data = self._get_proxy_data(app)
        cursor = proxy_data.query_items(
            'SELECT * FROM c WHERE ARRAY_CONTAINS(@L, c.commitId)',
            parameters=[
                dict(name='@L', value=commit_ids)
            ],
            enable_cross_partition_query=True,
        )
        return [DataRecord.parse_obj(item) for item in cursor]


    def batch_upsert_data_items(self, app: str, data_items: List[DataRecord]):
        raise NotImplementedError
    

    def batch_delete_data_items(self, app: str, data_items: List[DataRecord]):
        raise NotImplementedError
    
    
    
##
## Async Cosmos Proxy
##

class AsyncCosmosDataAccessProxy(DataAccessProxy):
    def __init__(
        self,
        client: aio.CosmosClient,
        loop: AbstractEventLoop,
        lock: RLock,
        dbname: str = DATABASE_DYNCONF,
    ):
        ''' Args:
                client: async cosmos client.
                loop: event loop bind to client.
                lock: lock for event loop.
                dbname: database name. Default: "dynconfig".
        '''
        self._dbname = dbname
        self._client = client
        self._loop = loop
        self._lock = lock
        
        self._proxy_db = self._client.get_database_client(self._dbname)
        self._proxy_meta = self._proxy_db.get_container_client(CONTAINER_META)
        self._proxy_commit = self._proxy_db.get_container_client(CONTAINER_COMMIT)
        self._proxy_data_dict: Dict[str, aio.ContainerProxy] = {}
    
    
    @staticmethod
    def from_cosmos_url(
        url: str,
        credential: Union[str, Dict[str, str], TokenCredential],
        dbname: str = DATABASE_DYNCONF,
    ) -> AsyncCosmosDataAccessProxy:
        ''' Args:
                url: cosmos url.
                credential: token with access. Need to use the async series of token objects, 
                            like `azure.identity.aio.ManagedIdentityCredential`.
                dbname: database name. Default: "dynconfig".
        '''
        client = aio.CosmosClient(url, credential)
        loop = asyncio.get_event_loop()
        lock = RLock()
        return AsyncCosmosDataAccessProxy(client, loop=loop, lock=lock, dbname=dbname)
        
    
    def close(self):
        ''' aio.CosmosClient requires to be closed after finished.
        '''
        with self._lock:
            self._loop.run_until_complete(self._client.close())

    
    def _get_data_proxy(self, app: str) -> aio.ContainerProxy:
        proxy_data = self._proxy_data_dict.get(app)
        if proxy_data is not None:
            return proxy_data
        
        # get container proxy
        assert app not in [CONTAINER_META, CONTAINER_COMMIT]
        proxy_data = self._proxy_db.get_container_client(app)

        # set
        self._proxy_data_dict[app] = proxy_data
        return proxy_data
        

    def try_get_meta(self, app: str) -> Optional[MetaRecord]:
        with self._lock:
            try:
                resp = self._loop.run_until_complete(
                    self._proxy_meta.read_item(app, partition_key=app)
                )
                return MetaRecord.parse_obj(resp)
            except CosmosResourceNotFoundError:
                return None

    
    def create_meta(self, meta: MetaRecord) -> MetaRecord:
        with self._lock:
            try:
                resp = self._loop.run_until_complete(
                    self._proxy_meta.create_item(meta.dict(exclude=FIELDS_EXCLUDE))
                )
                return MetaRecord.parse_obj(resp)
            except CosmosResourceExistsError as ex:
                raise ItemExistsException(f'meta "{meta.id}" already exists') from ex
            
    
    def update_meta_if_not_modified(self, meta: MetaRecord, etag: str) -> Optional[MetaRecord]:
        with self._lock:
            try:
                resp = self._loop.run_until_complete(
                    self._proxy_meta.replace_item(
                        meta.id,
                        meta.dict(exclude=FIELDS_EXCLUDE),
                        etag=etag,
                        match_condition=MatchConditions.IfNotModified,
                    )
                )
                return MetaRecord.parse_obj(resp)
            except CosmosAccessConditionFailedError:
                return None
        

    def create_commit(self, commit: CommitRecord) -> CommitRecord:
        with self._lock:
            try:
                resp = self._loop.run_until_complete(
                    self._proxy_commit.create_item(commit.dict(exclude=FIELDS_EXCLUDE))
                )
                return CommitRecord.parse_obj(resp)
            except CosmosResourceExistsError as ex:
                raise ItemExistsException(f'commit "{commit.id}" already exists') from ex
        
    
    def update_commit(self, commit: CommitRecord) -> CommitRecord:
        with self._lock:
            resp = self._loop.run_until_complete(
                self._proxy_commit.replace_item(commit.id, commit.dict(exclude=FIELDS_EXCLUDE))
            )
            return CommitRecord.parse_obj(resp)
        
    
    def batch_delete_commits(self, commits: List[CommitRecord]):
        with self._lock:
            self._loop.run_until_complete(
                self._batch_delete_commits(commits)
            )

    
    async def _batch_delete_commits(self, commits: List[CommitRecord]):
        async def delete_commit(c: CommitRecord):
            try:
                await self._proxy_commit.delete_item(c.id, partition_key=c.app)
            except CosmosResourceNotFoundError:
                pass
        await run_async_batch(
            delete_commit, 
            [Call(c) for c in commits],
        )


    def query_commits_by_seq_range(self, app: str, seq_begin: int, seq_end: int) -> List[CommitRecord]:
        with self._lock:
            return self._loop.run_until_complete(
                self._query_commits_by_seq_range(app, seq_begin, seq_end)
            )
            

    async def _query_commits_by_seq_range(self, app: str, seq_begin: int, seq_end: int) -> List[CommitRecord]:
        cursor = self._proxy_commit.query_items(
            'SELECT * FROM c WHERE c.seq between @seqBegin and @seqEnd',
            parameters=[
                dict(name="@seqBegin", value=seq_begin),
                dict(name="@seqEnd", value=seq_end),
            ],
            partition_key=app,
        )
        commits: List[CommitRecord] = []
        async for item in cursor:
            commits.append(CommitRecord.parse_obj(item))
        return commits
        
    
    def query_data_items_by_commit_ids(self, app: str, commit_ids: List[str]) -> List[DataRecord]:
        with self._lock:
            return self._loop.run_until_complete(
                self._query_data_items_by_commits(app, commit_ids)
            )


    async def _query_data_items_by_commits(self, app: str, commit_ids: List[str]) -> List[DataRecord]:
        proxy_data = self._get_data_proxy(app)
        cursor = proxy_data.query_items(
            'SELECT * FROM c WHERE ARRAY_CONTAINS(@L, c.commitId)',
            parameters=[
                dict(name='@L', value=commit_ids)
            ],
            partition_key=None,
        )
        drs: List[DataRecord] = []
        async for item in cursor:
            d = DataRecord.parse_obj(item)
            drs.append(d)
        return drs


    def batch_upsert_data_items(self, app: str, data_items: List[DataRecord]):
        with self._lock:
            self._loop.run_until_complete(
                self._batch_upsert_data_items(app, data_items)
            )
    
    
    async def _batch_upsert_data_items(self, app: str, data_items: List[DataRecord]):
        proxy_data = self._get_data_proxy(app)
        await run_async_batch(
            proxy_data.upsert_item,
            [Call(d.dict(exclude=FIELDS_EXCLUDE)) for d in data_items],
        )


    def batch_delete_data_items(self, app: str, data_items: List[DataRecord]):
        with self._lock:
            self._loop.run_until_complete(
                self._batch_delete_data_items(app, data_items)
            )
    
    
    async def _batch_delete_data_items(self, app: str, data_items: List[DataRecord]):
        proxy_data = self._get_data_proxy(app)
        async def delete_data_item(d: DataRecord):
            try:
                await proxy_data.delete_item(d.id, partition_key=d.key)
            except CosmosResourceNotFoundError:
                pass
        await run_async_batch(
            delete_data_item, 
            [Call(d) for d in data_items],
        )





from __future__ import annotations
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
import heapq
import logging
from threading import Condition, RLock, Thread
import time
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from azure.core.credentials import TokenCredential
from azure.cosmos import CosmosClient
from pyraisdk import rlog
from .core import CommitChainManager
from .proxy import CosmosDataAccessProxy, DataAccessProxy


class ConfigMonitor(ABC):
    ''' Base class for Dynamic Config Monitors to inherit from.
    '''
    def __init__(
        self,
        url: str,
        credential: Union[str, Dict[str, str], TokenCredential],
        app: str,
        poll_interval: float = 10,
    ):
        ''' Constructor. Won't really run until 'start' called.
            Args:
                url: The URL of the Cosmos DB account.
                credential: Can be the account key, or a dictionary of resource tokens.
                app: Application name of dynamic config.
                poll_interval: Poll interval in seconds. Default 10.
        '''
        self.__url = url
        self.__credential = credential
        self.__app = app
        self.__poll_interval = poll_interval
        # init load status
        self.__init_load_done: bool = False
        self.__init_load_cond = Condition()
        self.__init_load_excepted: Optional[Exception] = None
    

    def _set_init_load_status(self, success: bool, excepted: Optional[Exception]):
        ''' Should only be called in `_MonitorManager`.
            Strictly speaking it is not a private function. Naming as private method is
            to remind monitor users not to call it directly.
        '''
        if self.__init_load_done:
            return
        with self.__init_load_cond:
            if success:
                self.__init_load_done = True
                self.__init_load_excepted = None
                self.__init_load_cond.notify_all()
            else:
                assert excepted is not None
                self.__init_load_excepted = excepted
                self.__init_load_cond.notify_all()


    def wait_init_load(self, timeout: Optional[float] = None):
        ''' Wait for initial loading '''
        with self.__init_load_cond:
            if self.__init_load_done:
                return
            elif self.__init_load_excepted is not None:
                raise self.__init_load_excepted
        
            notified = self.__init_load_cond.wait(timeout)
            if notified:
                if self.__init_load_done:
                    return
                elif self.__init_load_excepted is not None:
                    raise self.__init_load_excepted
                else:
                    raise Exception('unexpected error')
            else:
                raise TimeoutError('initial loading timeout')


    def start(self, blocking: bool = False, timeout: Optional[float] = None):
        ''' Start config monitor
            Args:
                blocking: If True, block until initial loading complete for fail;
                    If False, return immediately.
                timeout: When blocking is True and it blocks exceed 'timeout', will
                    raise the TimeoutError exception. None for no timeout limitation.
        '''
        _manager.start_config_monitor(
            self, self.__url, self.__credential, self.__app, self.__poll_interval
        )
        if blocking:
            self.wait_init_load(timeout)
        

    def close(self):
        ''' Stop config monitor '''
        _manager.stop_config_monitor(self)


    @abstractmethod
    def apply(self, updates: Dict[str, Any], deletions: Set[str]):
        ''' Monitor apply function.
        
            Abstract method which need to be implemented in subclass. It receives
            a changeset (include updates and deletions) of one/multiple commits.

            If `apply` fails, for any reason, it won't retry automatically. But the current
            changeset (updates & deletions) will be preserved. When the next changeset (new data) 
            is coming, the two changesets will be merged and then applied once. If it fails again,
            the merged changeset will be preserved, and wait for next data coming.
            
            Args:
                updates: key-value to be updated
                deletions: keys to be deleted
        '''
        pass



@dataclass
class _MonitorHandler:
    app: str
    monitor: ConfigMonitor
    poll_interval: float
    chainman: CommitChainManager
    next_pull_seq: int = 0
    next_apply_seq: int = 0
    next_ts: float = 0
    lock: RLock = field(default_factory=RLock)


class _MonitorManager:
    def __init__(self):
        self._handlers: Dict[int, _MonitorHandler] = {}
        self._handler_heap: List[Tuple[float, _MonitorHandler]] = []
        self._clients: Dict[str, CosmosClient] = {}
        self._cond = Condition()
        self._pool = ThreadPoolExecutor()
        self._worker: Optional[Thread] = None
        
    
    def start_config_monitor(
        self,
        monitor: ConfigMonitor,
        url: str,
        credential: Union[str, Dict[str, str], TokenCredential],
        app: str,
        poll_interval: float,
    ):
        # get cosmos client
        with self._cond:
            if url in self._clients:
                client = self._clients[url]
            else:
                client = CosmosClient(url, credential)
                self._clients[url] = client
        
        # start
        proxy = CosmosDataAccessProxy(client)
        self._start_config_monitor_with_access_proxy(monitor, proxy, app, poll_interval)


    def _start_config_monitor_with_access_proxy(
        self,
        monitor: ConfigMonitor,
        proxy: DataAccessProxy,
        app: str,
        poll_interval: float,
    ):
        with self._cond:
            # set handler
            chainman = CommitChainManager(proxy, app)
            handler = _MonitorHandler(app, monitor, poll_interval, chainman)
            monitor_id = id(monitor)
            if monitor_id in self._handlers:
                raise Exception(f'ConfigMonitor {monitor_id} already started')
            else:
                heapq.heappush(self._handler_heap, (handler.next_ts, handler))
                self._handlers[monitor_id] = handler

            # init worker
            if self._worker is None:
                self._worker = Thread(target=self._worker_run, daemon=True)
                self._worker.start()
            else:
                self._cond.notify_all()


    def stop_config_monitor(self, monitor: ConfigMonitor):
        with self._cond:
            monitor_id = id(monitor)
            if monitor_id in self._handlers:
                self._handlers.pop(monitor_id)


    def _worker_run(self):
        while True:
            try:
                self._worker_run_inner()
            except Exception as ex:
                if rlog._logger_initialized:
                    rlog.errorcf('', -1, ex, f'{EVENT_KEY_PREFIX}: unexpected error in manager worker')
                else:
                    logging.exception(f'{EVENT_KEY_PREFIX}: unexpected error in manager worker')
                time.sleep(3)


    def _worker_run_inner(self):
        with self._cond:
            while True:
                if len(self._handler_heap) == 0:
                    self._cond.wait()
                else:
                    # pop handler that is ready to execute next
                    _, handler = heapq.heappop(self._handler_heap)
                    monitor_id = id(handler.monitor)
                    notified = self._cond.wait(handler.next_ts - time.time())
                    
                    # new handler added
                    if notified:
                        heapq.heappush(self._handler_heap, (handler.next_ts, handler))
                        continue
                    
                    # current handler deleted
                    if monitor_id not in self._handlers:
                        continue
                    
                    # next ts
                    handler.next_ts = time.time() + handler.poll_interval
                    heapq.heappush(self._handler_heap, (handler.next_ts, handler))
                    
                    # pull & apply
                    self._pool.submit(self._pull_and_apply, handler)


    def _pull_and_apply(self, handler: _MonitorHandler):
        # If handler is already executing, skip current iteration.
        acquired = handler.lock.acquire(blocking=False)
        if not acquired:
            return
        try:
            # refresh chain
            handler.chainman.refresh_chain(enable_init=False)
            # no newer commits, skip
            if handler.next_pull_seq > handler.chainman.meta.topSeq:
                return
            # pull, from next_apply_seq
            pulled_data = handler.chainman.fetch_changeset_of_chain(handler.next_apply_seq)
            handler.next_pull_seq = handler.chainman.meta.topSeq + 1
            # apply
            handler.monitor.apply(pulled_data.updates, pulled_data.deletions)
            handler.next_apply_seq = handler.chainman.meta.topSeq + 1
            # init load success
            handler.monitor._set_init_load_status(True, None)
            
        except Exception as ex:
            self._handle_monitor_exception(handler, ex)
            
        finally:
            handler.lock.release()


    def _handle_monitor_exception(self, handler: _MonitorHandler, excepted: Exception):
            # init load status
            handler.monitor._set_init_load_status(False, excepted)
            # logging
            if rlog._logger_initialized:
                rlog.event(f'{EVENT_KEY_PREFIX}_MonitorProcessError', handler.app, 0)
                rlog.errorcf('', -1, excepted, f'{EVENT_KEY_PREFIX}: monitor process error, app: {handler.app}')
            else:
                logging.error(f'{EVENT_KEY_PREFIX}: monitor process error, app: {handler.app}', exc_info=excepted)



# logging key prefix for monitor
EVENT_KEY_PREFIX = 'pyraidynconf'

# singleton
_manager = _MonitorManager()





from __future__ import annotations
import json
from threading import RLock
import time
from typing import Any, Dict, List, Optional, Set
import uuid
from azure.cosmos import CosmosClient
from .proxy import (
    AsyncCosmosDataAccessProxy, 
    CosmosDataAccessProxy, 
    DataAccessProxy, 
    ItemExistsException,
)
from .common import (
    MAX_COMMIT_CHAIN_LENGTH,
    MAX_COMMIT_DURATION_IN_SECONDS,
    MAX_WRITE_CONFLICT_RETRY_COUNT,
    CommitRecord,
    DataRecord, 
    MetaRecord,
)


class ExternalSequenceError(Exception):
    pass


# ChangeSet is for data implicitly. Not for meta or commit.
class ChangeSet:
    def __init__(self):
        self._updates: Dict[str, Any] = {}
        self._deletions: Set[str] = set()

    @property
    def updates(self) -> Dict[str, Any]:
        return self._updates
    
    @property
    def deletions(self) -> Set[str]:
        return self._deletions

    def set(self, key: str, val: Any):
        self._updates[key] = val
        if key in self._deletions:
            self._deletions.remove(key)
            
    def delete(self, key: str):
        self._deletions.add(key)
        if key in self._updates:
            self._updates.pop(key)

    def is_empty(self) -> bool:
        return not self._updates and not self._deletions
    
    def __contains__(self, k: str):
        return k in self._updates or k in self._deletions
    
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ChangeSet):
            return self._updates == other._updates and self._deletions == other._deletions
        return False

    def apply_on(self, dataset: Dict[str, Any]):
        ''' Apply changeset on the input dataset.
        '''
        dataset.update(self._updates)
        for k in self._deletions:
            if k in dataset:
                dataset.pop(k)

    def update(self, sub: ChangeSet):
        for k, v in sub._updates.items():
            self.set(k, v)
        for k in sub._deletions:
            self.delete(k)

    @classmethod
    def make_from_datasets(cls, src: Dict[str, Any], dst: Dict[str, Any]) -> ChangeSet:
        ''' Compare 2 datasets, and get the changeset from `src` to `dst`.
        '''
        changeset = ChangeSet()
        for key in src.keys() | dst.keys():
            if key in src and key in dst:
                src_val_str = json.dumps(src[key], sort_keys=True)
                dst_val_str = json.dumps(dst[key], sort_keys=True)
                if src_val_str != dst_val_str:
                    changeset.set(key, dst[key])
            elif key in src:
                changeset.delete(key)
            elif key in dst:
                changeset.set(key, dst[key])
        return changeset
    

class CommitChainManager:
    def __init__(self, proxy: DataAccessProxy, app: str):
        self._proxy = proxy
        self._app = app
        self._meta: Optional[MetaRecord] = None
        self._commits: Dict[int, CommitRecord] = {}  # key: seq
    
    
    @property
    def meta(self) -> MetaRecord:
        if self._meta is None: 
            raise Exception('require refresh_chain first')
        return self._meta
    
    
    def get_sub_chain(self, seq_begin: Optional[int] = None, seq_end: Optional[int] = None) -> List[CommitRecord]:
        if seq_begin is None:
            seq_begin = self.meta.rootSeq
        if seq_end is None:
            seq_end = self.meta.topSeq
        return [self._commits[seq] for seq in range(seq_begin, seq_end + 1)]
        
        
    def get_top_commit(self) -> CommitRecord:
        return self._commits[self.meta.topSeq]
    
    
    def refresh_chain(self, enable_init: bool = True):
        ''' Update (or initialize) the commit chain to latest.
            Args:
                enable_init: Indicates whether to initialize meta record when it doesn't exist.
                             Will raise Exception when it's False and meta not found.
        '''
        # try to get latest meta
        # when meta not exist, try to create initial meta, if enable_init.
        meta = self._proxy.try_get_meta(self._app)
        if meta is None:
            if self._meta is not None:
                # Won't happen unless cosmosdb records are manually deleted.
                raise Exception(f'unexpected error to get meta record "{self._app}"')
            elif not enable_init:
                raise Exception(f'meta record {self._app} does not exist')
            else:
                # create meta
                try:
                    meta = self._proxy.create_meta(
                        MetaRecord(
                            id=self._app, 
                            topSeq=0,
                            rootSeq=1,
                            topCommitId='',
                            _etag='',
                            _ts=0,
                        )
                    )
                except ItemExistsException:
                    # already exists, try get again
                    meta = self._proxy.try_get_meta(self._app)
                    if meta is None:
                        raise Exception(f'unexpected error for existed meta record {self._app}')
        
        # topSeq check for robustness, expect always meet.
        # But if somehow happend (can't imagine how), should raise out immediately.
        if self._meta is not None:
            assert self._meta.topSeq <= meta.topSeq

        # seq range of sub chain to fetch
        seq_begin = meta.rootSeq if self._meta is None else self._meta.topSeq + 1
        seq_end = meta.topSeq
        if seq_begin > seq_end:
            self._meta = meta
            return
        
        # query commit records and build chain
        commits = self._proxy.query_commits_by_seq_range(self._app, seq_begin, seq_end)
        chain = self.build_sub_commit_chain(commits, seq_begin, seq_end, meta.topCommitId)
        if self._meta is not None:
            # validate the join point of the chain
            if chain[0].parentId != self._meta.topCommitId or chain[0].seq != self._meta.topSeq + 1:
                raise Exception(f'invalid commit {self._meta.topCommitId} at seq {self._meta.topSeq}')

        # update
        self._commits.update({c.seq: c for c in chain})
        self._meta = meta


    def create_commit(self, ext_seq: int) -> CommitRecord:
        commit_req = CommitRecord(
            id=str(uuid.uuid4()),
            app=self._app,
            seq=self.meta.topSeq + 1,
            parentId=self.meta.topCommitId,
            extSeq=ext_seq,
            _etag='',
            _ts=0,
        )
        return self._proxy.create_commit(commit_req)


    def submit_change(self, commit: CommitRecord, squash: bool = False) -> bool:
        ''' Try to submit commit.
            Returns:
                True: success
                False: indicates etag conflict
        '''
        # validate commit not submitted, for robustness
        # assume it was success, but we got an error caused by any reason
        sub_chain = self.get_sub_chain(seq_begin=commit.seq)
        if commit.id in (c.id for c in sub_chain):
            raise Exception(f'commit {commit.id} already submitted')

        # update commit
        if commit.seq != self.meta.topSeq + 1:
            commit.seq = self.meta.topSeq + 1
            commit.parentId = self.meta.topCommitId
            self._proxy.update_commit(commit)
        
        # update meta from commit
        meta_req = self.meta.copy()
        meta_req.topSeq = commit.seq
        meta_req.topCommitId = commit.id
        if squash:
            meta_req.rootSeq = meta_req.topSeq
        
        meta_resp = self._proxy.update_meta_if_not_modified(meta_req, meta_req.etag)
        return meta_resp is not None


    # This method does not mutate the state of chain manager.
    def fetch_changeset_of_chain(self, from_seq: int) -> ChangeSet:
        ''' Get the changeset from `from_seq` to `topSeq`.
            `from_seq` == 0 indicates it's in full mode.
        '''
        seq_begin = self.meta.rootSeq if from_seq == 0 else from_seq
        seq_end = self.meta.topSeq
        if seq_begin > seq_end:
            return ChangeSet()

        # get commit chain
        sub_chain: List[CommitRecord] = self.get_sub_chain(seq_begin, seq_end)
        id_commits: Dict[str, CommitRecord] = {c.id: c for c in sub_chain}
        
        # data query
        data_items = self._proxy.query_data_items_by_commit_ids(self._app, [c.id for c in sub_chain])

        # build changeset
        key_item: Dict[str, DataRecord] = {}
        key_seq: Dict[str, int] = {}
        for ditem in data_items:
            commit = id_commits.get(ditem.commitId)
            assert commit is not None
            if commit.seq > key_seq.get(ditem.key, 0):
                key_item[ditem.key] = ditem
                key_seq[ditem.key] = commit.seq
                
        changeset = ChangeSet()
        for ditem in key_item.values():
            if ditem.deleted:
                changeset.delete(ditem.key)
            else:
                changeset.set(ditem.key, ditem.value)

        return changeset


    def put_changeset_into_commit(self, commit: CommitRecord, changeset: ChangeSet):
        if changeset.is_empty():
            return
        
        # build data records
        data_items: List[DataRecord] = []
        for k, v in changeset.updates.items():
            ditem = DataRecord(
                id=f'{k}:{commit.id}',
                commitId=commit.id,
                key=k,
                value=v,
                deleted=False,
                _etag='',   # placeholder; set by Cosmos DB server
                _ts=0,
            )
            data_items.append(ditem)
        
        for k in changeset.deletions:
            ditem = DataRecord(
                id=f'{k}:{commit.id}',
                commitId=commit.id,
                key=k,
                value=None,
                deleted=True,
                _etag='',
                _ts=0,
            )
            data_items.append(ditem)
        
        # upsert items
        self._proxy.batch_upsert_data_items(self._app, data_items)


    def validate_ext_seq_inc(self, ext_seq: int):
        ''' Verify the incrementality of the external sequence
        '''
        if self.meta.topSeq == 0 or ext_seq < 0:
            return
        commit = self._commits[self.meta.topSeq]
        if ext_seq <= commit.extSeq:
            raise ExternalSequenceError(f'external seq violate incremental constraint: {commit.extSeq} to {ext_seq}') 


    @staticmethod
    def build_sub_commit_chain(
        commits: List[CommitRecord], seq_begin: int, seq_end: int, end_commit_id: str
    ) -> List[CommitRecord]:
        ''' Build commit chain.
            Return commit records in order from `seq_begin` to `seq_end`.
        '''
        assert seq_begin >= 1
        commitdict = {c.id: c for c in commits}
        chain = []
        commit_id = end_commit_id
        # After the loop completes, all transient commit records will be removed.
        for seq in range(seq_end, seq_begin - 1, -1):
            commit = commitdict.get(commit_id)
            if commit is None or commit.seq != seq:
                # Not expected error point.
                # Only if someone deliberately messes with the data.
                raise Exception(f'invalid commit {commit_id} at seq {seq}')
            chain.append(commit)
            commit_id = commit.parentId
        chain.reverse()
        return chain


class DynamicConfigWriter:
    ''' It should be created with `AsyncCosmosDataAccessProxy` in Prod.
    
        Csomos async client has high write performance, but there are more
        restrictive conditions to use it. Mainly they are:
        1. It's hard to be used in a multi-threaded environment. You have 
           to face unexpected errors.
        2. Cosmos async client will be bind to the event loop after the 
           first execution, and unable to replace it again.
    '''
    
    def __init__(
        self,
        proxy: DataAccessProxy,
        app: str,
    ):
        self._chainman = CommitChainManager(proxy, app)
        self._lock = RLock()
    
    
    @staticmethod
    def make(proxy: AsyncCosmosDataAccessProxy, app: str):
        ''' Make a writer with `AsyncCosmosDataAccessProxy`.
            Should use this instead of `__init__` in Prod.
        '''
        return DynamicConfigWriter(proxy, app)


    def reconcile(self, dataset: Dict[str, Any], ext_seq: Optional[int] = None) -> ChangeSet:
        ''' Make remote dataset consistent with input `dataset`. 
            Success of reconcile guarantee at that moment remote dataset
            is completely same with local input `dataset`.
            
            Args:
                dataset: Local full dataset to sync to remote.
                ext_seq: An extra external sequence. For provided values, require
                        them to be incremental in commit chain.
            
            Returns:
                The changeset updated to remote in new commit.
        '''
        with self._lock:
            return self._reconcile(dataset, ext_seq)
    
    
    def gc(self):
        ''' Garbage collect.
            Can be executed at any time. Include 2 steps:
            1. sqaush commit chain`
            2. collect commit garbage
        '''
        with self._lock:
            self._squash_commit_chain()
            self._collect_commit_garbage()
    
    
    def _reconcile(self, dataset: Dict[str, Any], ext_seq: Optional[int] = None) -> ChangeSet:
        # ext_seq
        if ext_seq is None:
            ext_seq = -1
        else:
            assert ext_seq >= 0

        # initialize
        next_pull_seq = 0
        remote_dataset = {}
        commit_changeset = ChangeSet()
        commit: CommitRecord = None  # type: ignore
        ts_start = time.time()
        
        for trial in range(MAX_WRITE_CONFLICT_RETRY_COUNT):
            self._chainman.refresh_chain()

            # has more remote data
            if next_pull_seq <= self._chainman.meta.topSeq:
                # get data
                pulled_data = self._chainman.fetch_changeset_of_chain(next_pull_seq)
                next_pull_seq = self._chainman.meta.topSeq + 1
                pulled_data.apply_on(remote_dataset)

                # get delta
                if trial == 0:
                    # full delta, for first time
                    delta = ChangeSet.make_from_datasets(remote_dataset, dataset)
                    if delta.is_empty():
                        # nothing to commit, success automatically
                        return commit_changeset
                else:
                    # incremental delta (save cost of value compare):
                    # 1. only in scope of keys in `pulled_data`
                    # 2. exclude Keys already in `commit_changeset`
                    delta = ChangeSet.make_from_datasets(
                        {k: v for k, v in remote_dataset.items() if k in pulled_data and k not in commit_changeset}, 
                        {k: v for k, v in dataset.items()        if k in pulled_data and k not in commit_changeset}, 
                    )

                # check ext seq
                # It is intentional to put it here, other than after refresh_chain.
                # The purpose is to make it possible to call `reconcile` with exact same 
                # parameters repeatly.
                self._chainman.validate_ext_seq_inc(ext_seq)
                
                # write delta
                if commit is None:
                    commit = self._chainman.create_commit(ext_seq)
                self._chainman.put_changeset_into_commit(commit, delta)
                commit_changeset.update(delta)
            
            # max duration check
            if time.time() - ts_start > MAX_COMMIT_DURATION_IN_SECONDS:
                raise Exception(f'timeout for commit duration')
            
            # try submit
            success = self._chainman.submit_change(commit)
            if success:
                return commit_changeset
                
        raise Exception('exceed max write conflict retries')

    
    def _squash_commit_chain(
        self, 
        max_commit_chain_length: int = MAX_COMMIT_CHAIN_LENGTH,
    ):
        ''' To avoid the commit chain being too long, need to quash the long chain.
            Read out the full dataset from remote, and submit the dataset in one
            commit to overwrite the old data.
        '''
        next_pull_seq = 0
        commit: CommitRecord = None  # type: ignore
        ts_start = time.time()
        
        for trial in range(MAX_WRITE_CONFLICT_RETRY_COUNT):
            self._chainman.refresh_chain(enable_init=False)
            # check chain length
            chain_len = self._chainman.meta.topSeq - self._chainman.meta.rootSeq + 1
            if chain_len <=  max_commit_chain_length:
                return
            
            if commit is None:
                # create commit record
                ext_seq = self._chainman.get_top_commit().extSeq
                commit = self._chainman.create_commit(ext_seq)
            else:
                # update extSeq
                commit.extSeq = self._chainman.get_top_commit().extSeq
                
            # has more remote data
            if next_pull_seq <= self._chainman.meta.topSeq:
                # get data
                pulled_data = self._chainman.fetch_changeset_of_chain(next_pull_seq)
                next_pull_seq = self._chainman.meta.topSeq + 1
                
                # remove deletions for first pull (full dataset, trail == 0)
                # the full dataset (ChangeSet) might still include meaningless deletions,
                # excluding them can help to reduce garbage data records in squash commit.
                if trial == 0:
                    pulled_data.deletions.clear()
                
                # write data
                self._chainman.put_changeset_into_commit(commit, pulled_data)
                
            # max duration check
            if time.time() - ts_start > MAX_COMMIT_DURATION_IN_SECONDS:
                raise Exception(f'timeout for commit duration')
            
            # try submit
            success = self._chainman.submit_change(commit, squash=True)
            if success:
                return
        
        # failed exceed max retry
        raise Exception('exceed max write conflict retries')
    

    def _collect_commit_garbage(
        self,
        garbage_retention_duration: float = MAX_COMMIT_DURATION_IN_SECONDS * 2,
    ):
        ''' Failed commits produce garbage. Need regular cleaning.
            Since can't distinguish garbage and ongoing commits easily, only
            clear commits meeting all following conditions:
            1. Not in chain
            2. seq <= topSeq
            3. Inactive for long enough (12 hours)
        '''
        self._chainman.refresh_chain(enable_init=False)
        meta = self._chainman.meta
        
        # build full commit chain
        commits_all = self._chainman._proxy.query_commits_by_seq_range(
            self._chainman._app, 1, meta.topSeq)
        
        chain = self._chainman.build_sub_commit_chain(
            commits_all, 1, meta.topSeq, meta.topCommitId)
        
        commits_in_chain = {c.id: c for c in chain}
        
        # get garbage commits to be deleted
        now = time.time()
        garbage_commits: List[CommitRecord] = []
        for commit in commits_all:
            # in chain
            if commit.id in commits_in_chain:
                continue
            # seq out of scope
            if commit.seq > meta.topSeq:
                continue
            # inactive not long enough
            if commit.ts > now - garbage_retention_duration:
                continue
            # put in garbage list
            garbage_commits.append(commit)

        # no garbage
        if not garbage_commits:
            return
        # get garbage data records
        garbage_data_items = self._chainman._proxy.query_data_items_by_commit_ids(
            self._chainman._app, [c.id for c in garbage_commits],
        )
        # delete garbage data records
        self._chainman._proxy.batch_delete_data_items(self._chainman._app, garbage_data_items)
        # delete garbage commit records
        self._chainman._proxy.batch_delete_commits(garbage_commits)



class DynamicConfigConsumer:
    ''' It should be created with `CosmosDataAccessProxy` in Prod.
    '''    
    def __init__(
        self,
        proxy: DataAccessProxy,
        app: str,
    ):
        self._chainman = CommitChainManager(proxy, app)
        self._lock = RLock()
        self._next_pull_seq = 0
        

    @staticmethod
    def from_cosmos_client(client: CosmosClient, app: str) -> DynamicConfigConsumer:
        ''' Create dynamic config consumer with cosmos client.
        
            Args:
                client: cosmos client
                app: application name
                
            Returns:
                dynamic config consumer
        '''
        proxy = CosmosDataAccessProxy(client)
        return DynamicConfigConsumer(proxy, app)


    def pull(self) -> Optional[ChangeSet]:
        ''' Pull changesets incrementally.
            First time will get the full dataset.
            Returns:
                None if no newer commits; or will get the changeset.
        '''
        with self._lock:
            self._chainman.refresh_chain(enable_init=False)
            if self._next_pull_seq <= self._chainman.meta.topSeq:
                pulled_data = self._chainman.fetch_changeset_of_chain(self._next_pull_seq)
                self._next_pull_seq = self._chainman.meta.topSeq + 1
                return pulled_data
            else:
                return None
    
    
    def reset(self):
        ''' Reset the "_next_pull_seq".
            Will pull full dataset for next time.
        '''
        with self._lock:
            self._next_pull_seq = 0





from pydantic import BaseModel, Field
from typing import Any


DATABASE_DYNCONF = "dynconfig"
CONTAINER_META = 'dynconf_meta'
CONTAINER_COMMIT= 'dynconf_commit'

PARTITION_KEY_META = 'id'
PARTITION_KEY_COMMIT = 'app'
PARTITION_KEY_DATA = 'key'

MAX_WRITE_CONFLICT_RETRY_COUNT = 20
MAX_COMMIT_DURATION_IN_SECONDS = 3600 * 6
MAX_COMMIT_CHAIN_LENGTH = 20


# logical primary key: id
class MetaRecord(BaseModel):
    # [partition key]
    # format: f'{app}'
    id: str
    # sequence number of top commit in chain
    # initial value: 0
    topSeq: int
    # sequence number of root commit in chain
    # initial value: 1
    rootSeq: int
    # commit id of top commit in chain
    topCommitId: str
    # cosmos provided
    # intentionally not using default values (for _etag & _ts). Want to perform strict 
    # checks on the record from cosmos to ensure that all the fields are correctly provided.
    # same for `CommitRecord` and `DataRecord`.
    etag: str = Field(alias='_etag')
    ts: int = Field(alias='_ts')


# logical primary key: id
class CommitRecord(BaseModel):
    # commit id
    # random generated id, like `uuid4()`
    id: str
    # [partition key]
    # application name
    app: str
    # sequence number of commit in chain
    seq: int
    # commit id of parent commit in chain
    parentId: str
    # external incremental sequence number
    # negative number (like -1) will reset incrementality
    extSeq: int
    # cosmos provided
    etag: str = Field(alias='_etag')
    ts: int = Field(alias='_ts')
    

# logical primary key: commitId + key
class DataRecord(BaseModel):
    # format: f'{key}:{commitId}'
    id: str
    # the commit this data record belongs to
    commitId: str
    # [partition key]
    # key of config item
    key: str
    # value of config item
    value: Any
    # indicates this is a delete operation on the key
    # value will be ignored when deleted is True
    deleted: bool
    # cosmos provided
    etag: str = Field(alias='_etag')
    ts: int = Field(alias='_ts')



#
# A brief structure diagram of the algorithm
#
# +-----------------------+
# |       topSeq: 3       |
# |      rootSeq: 1       |
# | topCommitId: 2de34efe |
# +-----------------------+
#   :
#   : meta
#   v
# +-----------------------+         +--------------------+     +--------------------+
# |        seq: 3         |         | commitId: 2de34efe |     | commitId: 2de34efe |
# |     id: 2de34efe      |         |       key: a       |     |       key: c       |
# |  parentId: 511e31c4   |  data   |     value: 11      |     |    value: null     |
# |                       | ......> |   deleted: false   | ... |   deleted: true    |
# +-----------------------+         +--------------------+     +--------------------+
#   |
#   | commit chain
#   v
# +-----------------------+         +--------------------+     +--------------------+
# |        seq: 2         |         | commitId: 511e31c4 |     | commitId: 511e31c4 |
# |     id: 511e31c4      |         |       key: c       |     |       key: b       |
# |  parentId: f257feb8   |         |     value: 30      |     |     value: 20      |
# |                       | ......> |   deleted: false   | ... |   deleted: false   |
# +-----------------------+         +--------------------+     +--------------------+
#   |
#   |
#   v
# +-----------------------+         +--------------------+
# |        seq: 1         |         | commitId: f257feb8 |
# |     id: f257feb8      |         |       key: a       |
# |     parentId: ''      |         |     value: 10      |
# |                       | ......> |   deleted: false   |
# +-----------------------+         +--------------------+
#
#
# Code for diagram (tool: graph-easy)
# ```
# graph {flow: down}
# [m] {label: "topSeq: 3 \n rootSeq: 1 \n topCommitId: 2de34efe"}
# [c3] {label: "seq: 3 \n id: 2de34efe \n parentId: 511e31c4"}
# [c2] {label: "seq: 2 \n id: 511e31c4 \n parentId: f257feb8"}
# [c1] {label: "seq: 1 \n id: f257feb8 \n parentId: ''"}
# [c3.1] {label: "commitId: 2de34efe \n key: a \n value: 11 \n deleted: false"}
# [c3.2] {label: "commitId: 2de34efe \n key: c \n value: null \n deleted: true"}
# [c2.1] {label: "commitId: 511e31c4 \n key: c \n value: 30 \n deleted: false"}
# [c2.2] {label: "commitId: 511e31c4 \n key: b \n value: 20 \n deleted: false"}
# [c1.1] {label: "commitId: f257feb8 \n key: a \n value: 10 \n deleted: false"}
# [m]  .. meta ..> { flow: down } [c3]  -- commit chain --> { flow: down } [c2] -> { flow: down } [c1]
# [c3] .. data ..> { flow: left } [c3.1]  .. { flow: left } [c3.2]
# [c2] ..> { flow: left } [c2.1]  .. { flow: left } [c2.2]
# [c1] ..> { flow: left } [c1.1]
# ```
#
